# -*- coding: iso-8859-1 -*-
"""bung zu Konfidenzintervallen.

Vgl. Skript: "Schtztheorie II"
"""

import Tkinter, operator, math

totalDrawn = 13
thetaDiscPoints = 15
useGaussian = 0

def enumerate(seq):
	ind = 0
	for el in seq:
		yield ind, el
		ind += 1

def faculty(n):
	return reduce(operator.mul, range(1, n+1), 1)

def binomial(up, down):
	return faculty(up)/(faculty(down)*faculty(up-down))

def bernoulli(p, n, N):
	n = int(n)
	return binomial(N, n)*p**n*(1-p)**(N-n)

def gauss(x, mu, sigma):
	return 1/math.sqrt(2*3.14159)/sigma*\
			math.exp(-(x-mu)**2/2/sigma**2)

def sum(seq):
	return reduce(operator.add, seq, 0)

def normalize(seq):
	total = float(sum(seq))
	return [item/total for item in seq]

def getGaussLine(nPoints, interval, mu, sigma, gaussCache={}):
	lineKey = (mu, sigma)
	if not gaussCache.has_key(lineKey):
		discWidth = (interval[1]-interval[0])/float(nPoints)
		scale = lambda ind: interval[0]+discWidth*(ind+0.5)
		thisLine = [gauss(scale(i), mu, sigma) for i in range(nPoints)]
		gaussCache[lineKey] = normalize(thisLine)
	return gaussCache[lineKey]

def sillyGauss(index, mu, sigma, nPoints, interval):
	return getGaussLine(nPoints, interval, mu, sigma)[int(index)]


labelFont=("Helvetica", 8)


class TableauField(Tkinter.Label):
	colmasks = {
		0: "#%(intensity)02x0000",
		1: "#00%(intensity)02x00",
	}
	numKlasses = 2
	def __init__(self, parent, value, maxVal, *args, **kwargs):
		Tkinter.Label.__init__(self, parent, text=str(value), 
			foreground="#aaaaaa", font=labelFont, *args, **kwargs)
		self.changeCallbacks = []
		self.klass = 0
		self.setValue(value, maxVal)
		self._setBackground(maxVal)
		self.bind("<1>", self._clickCallback)

	def _callChangeCallbacks(self):
		for fun in self.changeCallbacks:
			fun()

	def _setBackground(self, maxVal=None):
		if maxVal is None:
			maxVal = self._maxValCache
		else:
			self._maxValCache = maxVal
		intensity = int(self.value/float(maxVal)*170)+85
		self.config(background=
			self.colmasks[self.klass]%{"intensity": intensity})
		self._callChangeCallbacks()

	def _clickCallback(self, ev=None):
		self.setKlass((self.klass+1)%self.numKlasses)

	def setValue(self, value, maxVal):
		self.value = value
		self.config(text="%0.4g"%value)
		self._setBackground(maxVal)

	def getValue(self):
		return self.value

	def setKlass(self, newKlass):
		self.klass = newKlass
		self._setBackground()

	def getKlass(self):
		return self.klass

	def addChangeCallback(self, callback):
		self.changeCallbacks.append(callback)


class TwoDTableau(Tkinter.Frame):
	def __init__(self, parent, numXFields, numYFields, *args, **kwargs):
		Tkinter.Frame.__init__(self, parent, *args, **kwargs)
		self._buildTableau(numXFields, numYFields)
		self.numXFields, self.numYFields = numXFields, numYFields
	
	def _buildTableau(self, numXFields, numYFields):
		self.fields = {}
		for y in range(numYFields):
			for x in range(numXFields):
				self.fields[x,y] = TableauField(self, 0, 1)
				self.fields[x,y].grid(row=y+1, column=x+1, 
					sticky=Tkinter.NE+Tkinter.SW)
		for y in range(numYFields):
			self.rowconfigure(y, weight=1, pad=2)
		for x in range(numYFields):
			self.columnconfigure(x, weight=1, pad=2)
	
	def setFieldVal(self, x, y, val, maxVal):
		self.fields[x,y].setValue(val, maxVal)

	def setFieldKlass(self, x, y, klass):
		self.fields[x,y].setKlass(klass)

	def iterIndices(self):
		for x in xrange(self.numXFields):
			for y in xrange(self.numYFields):
				yield (x,y)

	def iterRow(self, row):
		for x in xrange(self.numXFields):
			yield self.fields[x, row]

	def iterCol(self, col):
		for y in xrange(self.numYFields):
			yield self.fields[col, y]

	def iterFields(self):
		return self.fields.itervalues()


class Discrete2DFunc(dict):
	def __init__(self, numXPoints, numYPoints, xInterval, yInterval, func):
		xWidth = (xInterval[1]-xInterval[0])/float(numXPoints)
		self.xScale = lambda xInd: xInterval[0]+xWidth*(xInd+0.5)
		yWidth = (yInterval[1]-yInterval[0])/float(numYPoints)
		self.yScale = lambda yInd: yInterval[0]+yWidth*(yInd+0.5)

		self.numXPoints, self.numYPoints = numXPoints, numYPoints
		dict.__init__(self, [((x,y), func(self.xScale(x), self.yScale(y)))
			for x in range(numXPoints)
				for y in range(numYPoints)])

	def getNumX(self):
		return self.numXPoints
	
	def getNumY(self):
		return self.numYPoints

	def getNumPoints(self):
		return (self.numXPoints, self.numYPoints)

	def getMaximumVal(self):
		return max(self.values())


class ConfintInspector(TwoDTableau):
	def __init__(self, parent, discreteFunc, *args, **kwargs):
		self.discreteFunc = discreteFunc
		TwoDTableau.__init__(self, parent, discreteFunc.getNumX(),
			discreteFunc.getNumY(), *args, **kwargs)
		self._setVals()
		self._addAxes()
		self._addSums()
		self._addBindings()
	
	def _setVals(self):
		funcMax = self.discreteFunc.getMaximumVal()
		for x,y in self.iterIndices():
			self.setFieldVal(x, y, self.discreteFunc[x,y], funcMax)

	def _addAxes(self):
		# HACK ALERT: Ich nehme hier ein int() einer Achse, weil 4.5
		# Kugeln nicht gut sind.  Das kracht natrlich, wenn man die
		# Achsen vertauscht.  Au.
		xPoints, yPoints = self.discreteFunc.getNumPoints()
		for x in range(xPoints):
			Tkinter.Label(self, text="%d"%int(self.discreteFunc.xScale(x))
				).grid(row=0, column=x+1)
		for y in range(yPoints):
			Tkinter.Label(self, text="%1.3g"%self.discreteFunc.yScale(y)
				).grid(row=y+1, column=0)

	def _addSums(self):
		xPoints, yPoints = self.discreteFunc.getNumPoints()
		self.colSums = []
		for x in range(xPoints):
			self.colSums.append(Tkinter.Label(self, text="0"))
			#self.colSums[-1].grid(column=x+1, row=yPoints+2)
		self.rowSums = []
		for y in range(yPoints):
			self.rowSums.append(Tkinter.Label(self, text="0"))
			self.rowSums[-1].grid(row=y+1, column=xPoints+2)
		self.columnconfigure(xPoints+2, weight=0, minsize=60)

	def _addBindings(self):
		for col, row in self.iterIndices():
			self.fields[col, row].addChangeCallback(
				lambda row=row: self._recomputeRowSum(row))
			self.fields[col, row].addChangeCallback(
				lambda col=col: self._recomputeColSum(col))
			self.fields[col, row].bind("<3>", lambda ev, row=row:
				self._selectForRowSum(row))

	def _recomputeRowSum(self, row):
		rowsum = 0
		for field in self.iterRow(row):
			if field.getKlass()==1:
				rowsum += field.getValue()
		self.rowSums[row].config(text="%01.3g"%rowsum)
	
	def _recomputeColSum(self, col):
		colsum = 0
		for field in self.iterCol(col):
			if field.getKlass()==1:
				colsum += field.getValue()
		self.colSums[col].config(text="%01.3g"%colsum)

	def _selectForRowSum(self, row, targetSum=0.90):
		rowVals = []
		for ind, field in enumerate(self.iterRow(row)):
			rowVals.append((field.getValue(), ind))
			field.setKlass(0)
		targetIndices = []
		curSum = 0
		while curSum<targetSum and rowVals:
			maxTuple = max(rowVals)
			curSum += maxTuple[0]
			targetIndices.append(maxTuple[1])
			rowVals.remove(maxTuple)
		for x in targetIndices:
			self.setFieldKlass(x, row, 1)


if __name__=="__main__":
	if not useGaussian:
		myFunc = Discrete2DFunc(totalDrawn+1, thetaDiscPoints, 
			(0,totalDrawn+1), (0,1),
			lambda n, p, totalDrawn=totalDrawn: bernoulli(p, n, totalDrawn))
	else:
		myFunc = Discrete2DFunc(totalDrawn+1, thetaDiscPoints, 
			(0,totalDrawn+1), (0,1),
			lambda n, p, totalDrawn=totalDrawn: sillyGauss(n, p*(totalDrawn+1), 
				(totalDrawn+1)*p*(1-p)*2., totalDrawn+1, (0, totalDrawn+1)))
	root = Tkinter.Tk()
	t = ConfintInspector(root, myFunc)
	t.bind("q", lambda ev, t=t: t.quit())
	t.focus()
	t.pack(expand=1, fill=Tkinter.BOTH)
	root.mainloop()
